/*
	rpipic - A PIC Programmer for the Raspberry Pi
	Copyright (C) 2012  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/


#include "pic.hpp"

Pic::Pic(Board &board) : 
	Chip(board),
	m_vpptime(1000000),
	m_init_hold(1000000),
	m_cmd_delay(1000000),
	m_erase_time(50000000),
	m_end_program_time(50000000),
	m_program_time(10000000)
{
}

Pic::~Pic()
{
}

void Pic::InitPic(int64_t vpptime, int64_t init_hold, int64_t cmd_delay, int64_t erase_time, int64_t end_program_time, int64_t program_time)
{
	m_vpptime = vpptime;
	m_init_hold = init_hold;
	m_cmd_delay = cmd_delay;
	m_erase_time = erase_time;
	m_end_program_time = end_program_time;
	m_program_time = program_time;
}

void Pic::Reset()
{
	GetBoard().SetVppHigh(false);
	GetBoard().SetVddHigh(false);
	GetBoard().SetClockHigh(false);
	GetBoard().SetDataHigh(false);
	Wait(100000000);
	
	if (m_vpptime < 0){
		GetBoard().SetVddHigh(true);
		Wait(-m_vpptime);
		GetBoard().SetVppHigh(true);
		Wait(m_init_hold);
	}
	else {
		GetBoard().SetVppHigh(true);
		Wait(m_vpptime);
		GetBoard().SetVddHigh(true);
		Wait(m_init_hold);
	}
}

void Pic::IncrementAddress()
{
	Write(0x06, 6);
	Wait(m_cmd_delay);
}

void Pic::LoadConfiguration()
{
	Write(0x00, 6);
	Wait(m_cmd_delay);
	Write(0x0000, 16);
	Wait(m_cmd_delay);
}

void Pic::BeginEraseProgramCycle()
{
	Write(0x08, 6);
	Wait(m_cmd_delay);
	Wait(m_erase_time + m_program_time);
}

void Pic::EndEraseProgramCycle()
{
	Write(0x0E, 6);
	Wait(m_cmd_delay + m_end_program_time);
}

uint16_t Pic::ReadProgramWord(unsigned int bits)
{
	Write(0x04, 6);
	Wait(m_cmd_delay);
	uint16_t word = (Read(16) >> 1) & ((1 << bits) - 1);
	Wait(m_cmd_delay);
	return word;
}

uint8_t Pic::ReadDataWord(unsigned int bits)
{
	Write(0x05, 6);
	Wait(m_cmd_delay);
	uint8_t word = (Read(16) >> 1) & ((1 << bits) - 1);
	Wait(m_cmd_delay);
	return word;
}

void Pic::WriteProgramWord(uint16_t word, unsigned int bits)
{
	Write(0x02, 6);
	Wait(m_cmd_delay);
	Write((word & ((1 << bits) - 1)) << 1, 16);
	Wait(m_cmd_delay);
}

void Pic::WriteDataWord(uint16_t word, unsigned int bits)
{
	Write(0x03, 6);
	Wait(m_cmd_delay);
	Write((word & ((1 << bits) - 1)) << 1, 16);
	Wait(m_cmd_delay);
}

void Pic::PicReadProgram(Data &data, uint16_t pic_address, uint16_t data_address, uint16_t length,
	unsigned int bits, bool configuration, bool increment_address)
{
	Reset();
	if (increment_address){
		IncrementAddress();
	}
	if (configuration){
		LoadConfiguration();
	}
	for (unsigned int i = 0; i < pic_address; i++){
		IncrementAddress();
	}
	for (unsigned int i = 0; i < length; i++){
		uint16_t word = ReadProgramWord(bits);
		if (((1 << bits) - 1) != word){ //assume the default is all 1s
			data.SetData((data_address + i)*2, word & 0xFF);
			data.SetData((data_address + i)*2 + 1, word >> 8);
		}
		IncrementAddress();
	}
}

void Pic::PicReadData(Data &data, uint16_t pic_address, uint16_t data_address, uint16_t length)
{
	Reset();
	for (unsigned int i = 0; i < pic_address; i++){
		IncrementAddress();
	}
	for (unsigned int i = 0; i < length; i++){
		uint8_t word = ReadDataWord(8);
		if (0xFF != word){ //assume the default is all 1s
			data.SetData(data_address*2 + i, word & 0xFF);
		}
		IncrementAddress();
	}
}

void Pic::PicWriteProgram(Data &data, uint16_t pic_address, uint16_t data_address, uint16_t length,
	unsigned int bits, bool end_program, bool configuration, bool increment_address)
{
	Reset();
	if (increment_address){
		IncrementAddress();
	}
	if (configuration){
		LoadConfiguration();
	}
	for (unsigned int i = 0; i < pic_address; i++){
		IncrementAddress();
	}
	for (unsigned int i = 0; i < length; i++){
		uint16_t word = ((1 << bits) - 1);
		try {
			word = data.GetData((data_address + i)*2) | (data.GetData((data_address + i)*2 + 1) << 8);
		}
		catch (int){
			word = ((1 << bits) - 1);
		}
		WriteProgramWord(word, bits);
		BeginEraseProgramCycle();
		if (end_program){
			EndEraseProgramCycle();
		}
		uint16_t rword = ReadProgramWord(bits);
		if (rword != word){
			fprintf(stderr, "Failed to write program word at 0x%04X: wrote 0x%04X, read 0x%04X\n", i, word, rword);
			throw 0;
		}
		IncrementAddress();
	}
}

void Pic::PicWriteData(Data &data, uint16_t pic_address, uint16_t data_address, uint16_t length, bool end_program)
{
	Reset();
	for (unsigned int i = 0; i < pic_address; i++){
		IncrementAddress();
	}
	for (unsigned int i = 0; i < length; i++){
		uint16_t word = 0xFF;
		try {
			word = data.GetData(data_address*2 + i);
		}
		catch (int){
			word = 0xFF;
		}
		WriteDataWord(word, 8);
		BeginEraseProgramCycle();
		if (end_program){
			EndEraseProgramCycle();
		}
		uint16_t rword = ReadDataWord(8);
		if (rword != (word & 0xFF)){
			fprintf(stderr, "Failed to write data word at 0x%04X: wrote 0x%02X, read 0x%02X\n", i, word, rword);
			throw 0;
		}
		IncrementAddress();
	}
}

