/*
	rpipic - A PIC Programmer for the Raspberry Pi
	Copyright (C) 2012  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/


#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <argp.h>

#include "data.hpp"
#include "gpio.hpp"
#include "board.hpp"

#include "pic12.hpp"
#include "pic14.hpp"

const char *Chip::FormatBits(uint32_t bits, unsigned int n)
{
	unsigned int p = 0;
	static char buffer[256];
	for (unsigned int i = n; i > 0; i--){
		buffer[p] = (bits & (1 << (i-1))) ? '1' : '0';
		p++;
	}
	buffer[p] = 0;
	return buffer;
}

enum
{
	OPT_VDD=256,
	OPT_VPP,
	OPT_CLOCK,
	OPT_DATA_IN,
	OPT_DATA_OUT,
	OPT_FILE_GPIO,
	OPT_LIST_ChipIdS,
	OPT_DEVICE_ID,
	OPT_BOOT_CLOCK
};

enum ChipId
{
	CHIPID_PIC12,
	CHIPID_PIC14,
	CHIPID_UNKNOWN
};

enum Mode
{
	MODE_NONE,
	MODE_READ,
	MODE_PROGRAM,
	MODE_LIST_INFO,
	MODE_ERASE,
	MODE_BOOT
};

struct ChipIdMap
{
	const char *name;
	ChipId id;
};

class Arguments
{
	public:
		Arguments(int argc, char **argv) :
			m_continue(true),
			m_code(false),
			m_data(false),
			m_configuration(false),
			m_user_id(false),
			m_file_gpio(false),
			m_device_id(false),
			m_boot_clock(false),
			m_vdd(11),
			m_vpp(9),
			m_clk(8),
			m_data_in(24),
			m_data_out(25),
			m_ChipId(CHIPID_PIC14),
			m_mode(MODE_NONE),
			m_ChipName(0)
		{
			argp_parse (&s_argp_i, argc, argv, 0, 0, this);
		}
		
		bool GetContinue()
		{
			return m_continue;
		}
		
		bool GetCode()
		{
			return m_code;
		}
		
		bool GetData()
		{
			return m_data;
		}
		
		bool GetConfiguration()
		{
			return m_configuration;
		}
		
		bool GetUserId()
		{
			return m_user_id;
		}
		
		Mode GetMode()
		{
			return m_mode;
		}
		
		bool GetFileGpio()
		{
			return m_file_gpio;
		}
		
		bool GetDeviceId()
		{
			return m_device_id;
		}
		
		bool GetBootClock()
		{
			return m_boot_clock;
		}
		
		int GetVdd()
		{
			return m_vdd;
		}
		
		int GetVpp()
		{
			return m_vpp;
		}
		
		int GetClock()
		{
			return m_clk;
		}
		
		int GetDataOut()
		{
			return m_data_out;
		}
		
		int GetDataIn()
		{
			return m_data_in;
		}
		
		Chip *GetChip(Board &board)
		{
			switch (m_ChipId)
			{
				case CHIPID_PIC12:
					return new Pic12Bit(board, m_ChipName);
				case CHIPID_PIC14:
					return new Pic14Bit(board);
				default:
					return 0;
			}
		}
		
	private:
		static error_t parse_opt (int key, char *arg, struct argp_state *state)
		{
			Arguments *arguments = (Arguments *)state->input;
			return arguments->ParseOpt(key, arg, state);
		}
		
		error_t ParseOpt (int key, char *arg, struct argp_state *state)
		{
			switch (key)
			{
				case 'c':
					m_code = true;
					break;
				case 'd':
					m_data = true;
					break;
				case 'x':
					m_configuration = true;
					break;
				case 'u':
					m_user_id = true;
					break;
				case 'b':
					if (MODE_NONE != m_mode){
						argp_error(state, "Cannot specify -b with other mode option");
						break;
					}
					m_mode = MODE_BOOT;
					break;
				case 'p':
					if (MODE_NONE != m_mode){
						argp_error(state, "Cannot specify -p with other mode option");
						break;
					}
					m_mode = MODE_PROGRAM;
					break;
				case 'r':
					if (MODE_NONE != m_mode){
						argp_error(state, "Cannot specify -r with other mode option");
						break;
					}
					m_mode = MODE_READ;
					break;
				case 'l':
					if (MODE_NONE != m_mode){
						argp_error(state, "Cannot specify -l with other mode option");
						break;
					}
					m_mode = MODE_LIST_INFO;
					break;
				case 'e':
					if (MODE_NONE != m_mode){
						argp_error(state, "Cannot specify -e with other mode option");
						break;
					}
					m_mode = MODE_ERASE;
					break;
				case 'q':
					m_ChipName = arg;
					m_ChipId = CHIPID_UNKNOWN;
					for (unsigned int i = 0; s_ChipIdmap[i].name != 0; i++){
						if (!strcmp(arg, s_ChipIdmap[i].name)){
							m_ChipId = s_ChipIdmap[i].id;
						}
					}
					if (CHIPID_UNKNOWN == m_ChipId){
						argp_error(state, "Unknown Chip ID: %s", arg);
					}
					break;
				case OPT_VDD:
					m_vdd = atoi(arg);
					break;
				case OPT_VPP:
					m_vpp = atoi(arg);
					break;
				case OPT_CLOCK:
					m_clk = atoi(arg);
					break;
				case OPT_DATA_IN:
					m_data_in = atoi(arg);
					break;
				case OPT_DATA_OUT:
					m_data_out = atoi(arg);
					break;
				case OPT_FILE_GPIO:
					m_file_gpio = true;
					break;
				case OPT_LIST_ChipIdS:
					for (unsigned int i = 0; s_ChipIdmap[i].name != 0; i++){
						printf("%s\n", s_ChipIdmap[i].name);
					}
					m_continue = false;
					break;
				case OPT_DEVICE_ID:
					if ((MODE_LIST_INFO == m_mode) || (m_mode == MODE_NONE)){
						argp_error(state, "--device-id only valid with -r or -l");
					}
					m_device_id = true;
					break;
				case OPT_BOOT_CLOCK:
					if ((MODE_BOOT != m_mode) || (m_mode == MODE_NONE)){
						argp_error(state, "--boot-clock only valid with -b");
					}
					m_boot_clock = true;
					break;
				case ARGP_KEY_ARG:
					argp_usage (state);
					break;
				case ARGP_KEY_END:
					if (!m_continue){
					}
					else if (m_mode == MODE_NONE){
						argp_error(state, "Not doing anything because no mode specified");
					}
					else {
						if (CHIPID_UNKNOWN == m_ChipId){
							argp_error(state, "Must specify -q for -p, -r, -l, or -e");
						}
					}
					break;
				default:
					return ARGP_ERR_UNKNOWN;
			}
			return 0;
		}
		
		static ChipIdMap s_ChipIdmap[];
		static argp_option s_options[];
		static argp s_argp_i;
		
		bool m_continue;
		bool m_code;
		bool m_data;
		bool m_configuration;
		bool m_user_id;
		bool m_file_gpio;
		bool m_device_id;
		bool m_boot_clock;
		int m_vdd;
		int m_vpp;
		int m_clk;
		int m_data_in;
		int m_data_out;
		ChipId m_ChipId;
		Mode m_mode;
		const char *m_ChipName;
};

argp_option Arguments::s_options[] = {
	{"program", 'p', 0, 0, "Program mode"},
	{"read", 'r', 0, 0, "Read mode"},
	{"erase", 'e', 0, 0, "Erase mode - erases everything on the Chip"},
	{"list-info", 'l', 0, 0, "List mode - prints information from standard in"},
	{"boot", 'b', 0, 0, "Boot mode (starts the Chip in non-program/verify mode)"},
	{"boot-clock", OPT_BOOT_CLOCK, 0, 0, "Use external clock for boot"},
	{"code", 'c', 0, 0, "Program/read code"},
	{"data", 'd', 0, 0, "Program/read data"},
	{"configuration", 'x', 0, 0, "Program/read configuration word"},
	{"user-id", 'u', 0, 0, "Program/read user id (synonymous with -x if --list-info specified)"},
	{"chip", 'q', "CHIP", 0, "Specific ChipId (see --list-chips)"},
	{"list-chips", OPT_LIST_ChipIdS, 0, 0, "List chips"},
	{"device-id", OPT_DEVICE_ID, 0, 0, "Read device ID (synonymous with -x if --list-info specified)"},
	{"vdd", OPT_VDD, "VDD", 0, "GPIO pin for Vdd"},
	{"vpp", OPT_VPP, "VPP", 0, "GPIO pin for Vpp"},
	{"clk", OPT_CLOCK, "CLK", 0, "GPIO pin for clock"},
	{"data-in", OPT_DATA_IN, "DATAIN", 0, "GPIO pin for data input"},
	{"data-out", OPT_DATA_OUT, "DATAOUT", 0, "GPIO pin for data output"},
	{"file-gpio", OPT_FILE_GPIO, 0, 0, "Use /sys/class/gpio interface"},
	{0}
};
argp Arguments::s_argp_i = {s_options, parse_opt, 0, "RPiChip - Raspberry Pi Chip Programmer"};

ChipIdMap Arguments::s_ChipIdmap[] = 
{
	//Non-autodetecting chips
	//{"pic10f200", CHIPID_PIC12},
	{"pic16f54", CHIPID_PIC12},
	
	//Autodetecting chips
	{"pic14bit", CHIPID_PIC14},
	
	//Not found 
	{0, CHIPID_UNKNOWN}
};

void Main(Arguments &arguments)
{
	if (MODE_LIST_INFO == arguments.GetMode()){
		Hex data;
		NullBoard board;
		Chip *chip = arguments.GetChip(board);
		data.ReadFile(stdin);
		if (arguments.GetCode()){
		}
		if (arguments.GetData()){
		}
		if (arguments.GetConfiguration() || arguments.GetUserId() || arguments.GetDeviceId()){
			chip->PrintMeta(data);
		}
		delete chip;
	}
	else if (
			(MODE_READ == arguments.GetMode()) ||
			(MODE_PROGRAM == arguments.GetMode()) ||
			(MODE_ERASE == arguments.GetMode()) ||
			(MODE_BOOT == arguments.GetMode())
		){
		Gpio *gpio = arguments.GetFileGpio() ? (Gpio *)new FileGpio : (Gpio *)new DirectGpio;
		Breadboard *board = new Breadboard(*gpio,
			arguments.GetVdd(),
			arguments.GetVpp(),
			arguments.GetClock(),
			arguments.GetDataOut(),
			arguments.GetDataIn());
		
		Chip *chip = 0;
		if (MODE_BOOT != arguments.GetMode()){
			chip = arguments.GetChip(*board);
		}
		
		if (MODE_BOOT == arguments.GetMode()){
			board->SetVppHigh(false);
			board->SetVddHigh(false);
			board->SetClockHigh(false);
			board->SetDataHigh(false);
			Chip::Wait(100000000);
			board->SetVddHigh(true);
			
			if (arguments.GetBootClock()){
				printf("Entering infinite clock loop\n");
				while (true){
					board->SetClockHigh(true);
					Chip::Wait(1000);
					board->SetClockHigh(false);
					Chip::Wait(1000);
				}
			}
			else {
				printf("Press enter to exit\n");
				fgetc(stdin);
			}
		}
		else if (MODE_READ == arguments.GetMode()){
			Hex data;
			if (arguments.GetCode()){
				chip->ReadProgram(data);
			}
			if (arguments.GetData()){
				chip->ReadData(data);
			}
			if (arguments.GetConfiguration()){
				chip->ReadConfiguration(data);
			}
			if (arguments.GetUserId()){
				chip->ReadUserId(data);
			}
			if (arguments.GetDeviceId()){
				chip->ReadDeviceId(data);
			}
			data.WriteFile(stdout);
		}
		else if (MODE_PROGRAM == arguments.GetMode()){
			Hex data;
			data.ReadFile(stdin);
			if (arguments.GetCode()){
				chip->WriteProgram(data);
			}
			if (arguments.GetData()){
				chip->WriteData(data);
			}
			if (arguments.GetConfiguration()){
				chip->WriteConfiguration(data);
			}
			if (arguments.GetUserId()){
				chip->WriteUserId(data);
			}
		}
		else if (MODE_ERASE == arguments.GetMode()){
			chip->Erase();
		}

		delete chip;
		delete board;
		delete gpio;
	}
}

int main(int argc, char **argv)
{
	Arguments arguments(argc, argv);	
	
	if (arguments.GetContinue()){
		try {
			Main(arguments);
		}
		catch (int){}
	}
	return 0;
}

